"""
A module implementing wrapper for accessing certain Sysmetric device features
"""

from pymodbus.client import ModbusUdpClient as ModbusClient
from pymodbus.transaction import ModbusRtuFramer

UNIT = 0x1

class DeviceType:
    """
    Defines device type to be connected
    """
    Artic = 1
    Samson = 2


system_states = [
    'Unknown',
    'On',
    'Off',
    'Ambienting',
    'Stand By',
    'Start',
    'Power Up'
]

artic_registers = {
        "system_control_state": 1,
        "alarm": 3,
        "on_off": 9,
        "set_temperature": 10,
        "tip_temperature": 12,
}

samson_registers = {
        "system_control_state": 1,
        "alarm": 3,
        "on_off": 9,
        "set_temperature": 10,
        "tip_temperature": 16,
}

artic_error_descriptions = [
    'LCD Error',
    'Compressor Temperature Error',
    'Compressor Pressure Error',
    'Head Communication Error',
    'Tip1 Temperature Sensor Error',
    'T-Case Temperature Sensor Error',
    'Head Safety Over Heat Error',
    'LOCAL Manage while in InTEC',
    'Tips Difference Error',
    'FB Error',
    'Tip-FB Difference Error',
    'Backup Battery Low Voltage',
    'Lost I2C Control',
    'FAN Failure Error',
    'T-Case Alarm',
    'Tip-Case Difference Alarm',
    'Undefined Alarm 17',
    'Undefined Alarm 18',
    'Undefined Alarm 19',
    'Undefined Alarm 20',
    'Undefined Alarm 21',
    'Undefined Alarm 22',
    'Undefined Alarm 23',
    'Undefined Alarm 24',
    'Undefined Alarm 25',
    'Undefined Alarm 26',
    'Undefined Alarm 27',
    'Undefined Alarm 28',
    'Undefined Alarm 29',
    'Undefined Alarm 30',
    'Undefined Alarm 31',
    'Ethernet Error'
]

samson_error_descriptions = [
    'LCD Error',
    'Compressor Temperature Error',
    'Compressor Pressure Error',
    'Head Communication Error',
    'Tip1 Temperature Sensor Error',
    'Tip2 Temperature Sensor Error',
    'Tip3 Temperature Sensors Error',
    'Tip4 Temperature Sensors Error',
    'Tips Difference Error',
    'FB Error',
    'Tip-FB Difference Error',
    'Backup Battery Low Voltage',
    'Lost I2C Control/FB Temp Error',
    'FAN Failure Error',
    'LOCAL Manage while in InTEC',
    'Head Safety Over Heat Error',
    'Water Leak Detected',
    'Undefined Alarm 18',
    'Undefined Alarm 19',
    'Undefined Alarm 20',
    'Undefined Alarm 21',
    'Undefined Alarm 22',
    'Undefined Alarm 23',
    'Undefined Alarm 24',
    'Undefined Alarm 25',
    'Undefined Alarm 26',
    'Undefined Alarm 27',
    'Undefined Alarm 28',
    'Undefined Alarm 29',
    'Undefined Alarm 30',
    'Undefined Alarm 31',
    'Ethernet Error',
]


class ThermalAdapter:
    """
    Class wrapping access to the Sysmetric's devices via Modbus-UDP protocol
    """

    def __init__(self, dev_type, host):
        """
        Initializes the ThermalAdapter instance
        :param dev_type: the Sysmetric's device type: 1 - Artic, 2- Samson
        :param host: the device host IP address
        """
        self._client = ModbusClient(host, port=502, framer=ModbusRtuFramer, timeout=1)
        self._opened = True
        if dev_type == DeviceType.Artic:
            self._registers = artic_registers
            self._error_descriptions = artic_error_descriptions
        else:
            self._registers = samson_registers
            self._error_descriptions = samson_error_descriptions        
        
    def close(self):
        """ Closes the session to the Sysmetric's device"""
        if self._opened:
            self._client.close()
            self._opened = False

    def _check_connection(self):
        """
        Checks that  the connection is opened
        :raises: RuntimeError if connection to the device is closed
        """
        if not self._opened:
            raise RuntimeError('Connection to device was closed')

    def get_actual_temperature(self):
        """
        Gets actual measured temperature in degrees Celsius
        :return: the actual temperature
        :raises: RuntimeError if connection to the device is closed
        """
        self._check_connection()

        regs = self._client.read_holding_registers(self._registers['tip_temperature'],1,UNIT)
        temp = regs.registers[0]
        #if temp > (1 << 31):
        #   temp -= (1 << 32)  # convert big unsigned to negative
        """if (temp  <= 65535 and  temp >= 64836) :
           temp = int(temp - 2**16)"""
        if (temp  <= 65535 and  temp >= 64836):
           temp = (temp - 2**16)   
        temp /= 10
        return temp

    def get_requested_temperature(self):
        """
        Gets last requested measured temperature in degrees Celsius
        :return: the requested temperature
        :raises: RuntimeError if connection to the device is closed
        """
        self._check_connection()

        regs = self._client.read_holding_registers(self._registers['set_temperature'],1,UNIT)
        temp = regs.registers[0]
        """if (temp >  1 << 31):
        temp -= (1 << 32)  # convert big unsigned to negative"""
        if (temp  <= 65535 and  temp >= 64836):
           temp = (temp - 2**16)
        temp /= 10
        
        return temp
    
    def set_temperature(self, temperature):
        """
        Sets the temperature
        :param temperature: the temperature to be set in degrees Celsius
        :raises: RuntimeError if connection to the device is closed
        :raises: ValueError if requested temperature is not valid
        """
        self._check_connection()

        if temperature < -999 or temperature > 999:
            raise ValueError('Invalid temperature value') 
        set_temperature = int(temperature * 10)
        if  set_temperature < 0 :
            set_temperature += (1 << 16)  # convert signed int to unsigned
            """
            if  (set_temperature  <= 65535 and  set_temperature >= 64836) :
                 set_temperature = int(set_temperature + 2**16)"""
        self._client.write_registers(self._registers['set_temperature'], [set_temperature], UNIT)
            
    def _toggle_status(self):
        """
        Toggles the device ON/Off state
        """
        self._client.write_registers(self._registers['on_off'], [1], UNIT)
        
    def get_control_state(self):
        """
        Gets current device controls state
        :return: current device controls state
        :raises: RuntimeError if connection to the device is closed
        """
        self._check_connection()

        response = self._client.read_holding_registers(self._registers['system_control_state'], 1, UNIT)
        return response.registers[0]

    def get_control_state_description(self, state):
        """
        Gets the controls state description for the provided control state number
        :param state: the device state ID
        :return: The controls state description
        :raises: ValueError if provided state is not valid
        """
        if state < 1 or state >= len(system_states):
            raise ValueError('Invalid state {}'.format(state))

        return system_states[state]

    def turn_on(self):
        """
        Turns the device on
        :raises: RuntimeError if connection to the device is closed
        """
        self._check_connection()
        control_state = self.get_control_state()

        if control_state != 1:
            self._toggle_status()

    def turn_off(self):
        """
        Turns the device off
        :raises: RuntimeError if connection to the device is closed
        """
        self._check_connection()
        control_state = self.get_control_state()

        if control_state != 2:
            self._toggle_status()

    def _alarm_to_errors(self, alarms):
        """
        Returns the list of error id and description pairs for the alarms bits
        :param alarms: the bit mask of set alarms
        :return: The list of error id and description pairs
        """
        errors = []
        for i in range(0, 32):
            if alarms & (1 << i) != 0:
                errors.append((i, self._error_descriptions[i]))
                
        return errors

    def get_errors(self):
        """
        Returns the current list of error id and description pairs
        :return: The list of error id and description pairs
        """
        self._check_connection()

        response = self._client.read_holding_registers(self._registers['alarm'], 2, UNIT)
        alarms = response.registers[0] | (response.registers[1] << 16)

        return self._alarm_to_errors(alarms)

#---------------------------------------------------------------------------#
# Exported symbols
#---------------------------------------------------------------------------#
__all__ = [
    "ThermalAdapter", "DeviceType"
]
